@@ -3,9 +3,10 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (c)  2002-2020, University of Amsterdam
+    Copyright (c)  2002-2023, University of Amsterdam
                               VU University Amsterdam
                               CWI, Amsterdam
+                              SWI-Prolog Solutions b.v.
     All rights reserved.
 
     Redistribution and use in source and binary forms, with or without
@@ -68,7 +69,7 @@
 
 /** <module> HTTP client library
 
-This library defines http_open/3, which opens a  URL as a Prolog stream.
+This library defines http_open/3, which opens an URL as a Prolog stream.
 The functionality of the  library  can   be  extended  by  loading two
 additional modules that act as plugins:
 
@@ -96,58 +97,59 @@ additional modules that act as plugins:
 
 Here is a simple example to fetch a web-page:
 
-  ==
-  ?- http_open('http://www.google.com/search?q=prolog', In, []),
-     copy_stream_data(In, user_output),
-     close(In).
-  <!doctype html><head><title>prolog - Google Search</title><script>
-  ...
-  ==
+```
+?- http_open('http://www.google.com/search?q=prolog', In, []),
+   copy_stream_data(In, user_output),
+   close(In).
+<!doctype html><head><title>prolog - Google Search</title><script>
+...
+```
 
 The example below fetches the modification time of a web-page. Note that
-Modified is '' (the empty atom)  if   the  web-server does not provide a
+=|Modified|= is =|''|= (the empty atom) if the  web-server does not provide a
 time-stamp for the resource. See also parse_time/2.
 
-  ==
-  modified(URL, Stamp) :-
-          http_open(URL, In,
-                    [ method(head),
-                      header(last_modified, Modified)
-                    ]),
-          close(In),
-          Modified \== '',
-          parse_time(Modified, Stamp).
-  ==
+```
+modified(URL, Stamp) :-
+       http_open(URL, In,
+                 [ method(head),
+                   header(last_modified, Modified)
+                 ]),
+       close(In),
+       Modified \== '',
+       parse_time(Modified, Stamp).
+```
 
 Then next example uses Google search. It exploits library(uri) to manage
 URIs, library(sgml) to load  an  HTML   document  and  library(xpath) to
 navigate the parsed HTML. Note that  you   may  need to adjust the XPath
-queries if the data returned by Google changes.
+queries if the data returned by Google changes (this example indeed
+no longer works and currently fails at the first xpath/3 call)
 
-  ==
-  :- use_module(library(http/http_open)).
-  :- use_module(library(xpath)).
-  :- use_module(library(sgml)).
-  :- use_module(library(uri)).
+```
+:- use_module(library(http/http_open)).
+:- use_module(library(xpath)).
+:- use_module(library(sgml)).
+:- use_module(library(uri)).
 
-  google(For, Title, HREF) :-
-          uri_encoded(query_value, For, Encoded),
-          atom_concat('http://www.google.com/search?q=', Encoded, URL),
-          http_open(URL, In, []),
-          call_cleanup(
-              load_html(In, DOM, []),
-              close(In)),
-          xpath(DOM, //h3(@class=r), Result),
-          xpath(Result, //a(@href=HREF0, text), Title),
-          uri_components(HREF0, Components),
-          uri_data(search, Components, Query),
-          uri_query_components(Query, Parts),
-          memberchk(q=HREF, Parts).
-  ==
+google(For, Title, HREF) :-
+        uri_encoded(query_value, For, Encoded),
+        atom_concat('http://www.google.com/search?q=', Encoded, URL),
+        http_open(URL, In, []),
+        call_cleanup(
+            load_html(In, DOM, []),
+            close(In)),
+        xpath(DOM, //h3(@class=r), Result),
+        xpath(Result, //a(@href=HREF0, text), Title),
+        uri_components(HREF0, Components),
+        uri_data(search, Components, Query),
+        uri_query_components(Query, Parts),
+        memberchk(q=HREF, Parts).
+```
 
 An example query is below:
 
-==
+```
 ?- google(prolog, Title, HREF).
 Title = 'SWI-Prolog',
 HREF = 'http://www.swi-prolog.org/' ;
@@ -161,7 +163,7 @@ Title = 'Learn Prolog Now!',
 HREF = 'http://www.learnprolognow.org/' ;
 Title = 'Free Online Version - Learn Prolog
 ...
-==
+```
 
 @see load_html/3 and xpath/3 can be used to parse and navigate HTML
      documents.
@@ -189,6 +191,7 @@ Title = 'Free Online Version - Learn Prolog
                        final_url(-atom),
                        header(+atom, -atom),
                        headers(-list),
+                       raw_headers(-list(string)),
                        connection(+atom),
                        method(oneof([delete,get,put,head,post,patch,options])),
                        size(-integer),
@@ -278,10 +281,13 @@ user_agent('SWI-Prolog').
 %     AtomValue is unified to the empty atom ('').
 %
 %     * headers(-List)
-%     If provided, List is unified with  a list of Name(Value) pairs
-%     corresponding to fields in the reply   header.  Name and Value
-%     follow the same conventions  used   by  the header(Name,Value)
-%     option.
+%     If provided,  List is unified  with a list of  Name(Value) pairs
+%     corresponding to  fields in  the reply  header.  Name  and Value
+%     follow  the  same  conventions used  by  the  header(Name,Value)
+%     option.  A  pseudo header status_code(Code) is  added to provide
+%     the  HTTP status  as  an integer.   See also  raw_headers(-List)
+%     which  provides  the  entire   HTTP  reply  header  in  unparsed
+%     representation.
 %
 %     * method(+Method)
 %     One of =get= (default), =head=, =delete=, =post=,   =put=   or
@@ -309,6 +315,15 @@ user_agent('SWI-Prolog').
 %     to   ask   for    bytes    1000-1999,     use    the    option
 %     range(bytes(1000,1999))
 %
+%     * raw_encoding(+Encoding)
+%     Do not install a decoding filter for Encoding.  For example,
+%     using raw_encoding('applocation/gzip') the system will not
+%     decompress the stream if it is compressed using `gzip`.
+%
+%     * raw_headers(-Lines)
+%     Unify Lines with a list of strings that represents the complete
+%     reply header returned by the server.  See also headers(-List).
+%
 %     * redirect(+Boolean)
 %     If `false` (default `true`), do _not_ automatically redirect
 %     if a 3XX code is received.  Must be combined with
@@ -386,14 +401,14 @@ user_agent('SWI-Prolog').
 %               Note that values must *not* be quoted because the
 %               library inserts the required quotes.
 %
-%               ==
+%               ```
 %               http_open([ host('www.example.com'),
 %                           path('/my/path'),
 %                           search([ q='Hello world',
 %                                    lang=en
 %                                  ])
 %                         ])
-%               ==
+%               ```
 %
 %   @throws error(existence_error(url, Id),Context) is raised if the
 %   HTTP result code is not in the range 200..299. Context has the
@@ -472,8 +487,8 @@ try_http_proxy(Method, Parts, Stream, Options0) :-
               [ Host:Port, StreamPair ]),
         catch(send_rec_header(StreamPair, Stream, HostPort,
                               RequestURI, Parts, Options),
-              error(E,_),
-              keep_alive_error(E))
+              Error,
+              keep_alive_error(Error, StreamPair))
     ->  true
     ;   http:http_connection_over_proxy(Method, Parts, Host:Port,
                                         SocketStreamPair, Options, Options1),
@@ -632,6 +647,7 @@ guarded_send_rec_header(StreamPair, Stream, Host, RequestURI, Parts, Options) :-
                                     % read the reply header
     read_header(StreamPair, Parts, ReplyVersion, Code, Comment, Lines),
     update_cookies(Lines, Parts, Options),
+    reply_header(Lines, Options),
     do_open(ReplyVersion, Code, Comment, Lines, Options, Parts, Host,
             StreamPair, Stream).
 
@@ -808,9 +824,9 @@ do_open(Version, Code, _, Lines, Options, Parts, Host, In0, In) :-
     return_version(Options, Version),
     return_size(Options, Headers),
     return_fields(Options, Headers),
-    return_headers(Options, Headers),
+    return_headers(Options, [status_code(Code)|Headers]),
     consider_keep_alive(Lines, Parts, Host, In0, In1, Options),
-    transfer_encoding_filter(Lines, In1, In),
+    transfer_encoding_filter(Lines, In1, In, Options),
                                     % properly re-initialise the stream
     set_stream(In, file_name(URI)),
     set_stream(In, record_position(true)).
@@ -993,7 +1009,7 @@ return_final_url(Options) :-
 return_final_url(_).
 
 
-%!  transfer_encoding_filter(+Lines, +In0, -In) is det.
+%!  transfer_encoding_filter(+Lines, +In0, -In, +Options) is det.
 %
 %   Install filters depending on the transfer  encoding. If In0 is a
 %   stream-pair, we close the output   side. If transfer-encoding is
@@ -1002,19 +1018,23 @@ return_final_url(_).
 %   on  this.  Exceptions  to  this   are  content-types  for  which
 %   disable_encoding_filter/1 holds.
 
-transfer_encoding_filter(Lines, In0, In) :-
+transfer_encoding_filter(Lines, In0, In, Options) :-
     transfer_encoding(Lines, Encoding),
     !,
-    transfer_encoding_filter_(Encoding, In0, In).
-transfer_encoding_filter(Lines, In0, In) :-
+    transfer_encoding_filter_(Encoding, In0, In, Options).
+transfer_encoding_filter(Lines, In0, In, Options) :-
     content_encoding(Lines, Encoding),
     content_type(Lines, Type),
     \+ http:disable_encoding_filter(Type),
     !,
-    transfer_encoding_filter_(Encoding, In0, In).
-transfer_encoding_filter(_, In, In).
+    transfer_encoding_filter_(Encoding, In0, In, Options).
+transfer_encoding_filter(_, In, In, _Options).
 
-transfer_encoding_filter_(Encoding, In0, In) :-
+transfer_encoding_filter_(Encoding, In0, In, Options) :-
+    option(raw_encoding(Encoding), Options),
+    !,
+    In = In0.
+transfer_encoding_filter_(Encoding, In0, In, _Options) :-
     stream_pair(In0, In1, Out),
     (   nonvar(Out)
     ->  close(Out)
@@ -1225,6 +1245,18 @@ rest_(Atom, L, []) :-
     atom_codes(Atom, L).
 
 
+%!  reply_header(+Lines, +Options) is det.
+%
+%   Return the entire reply header as  a list of strings to the option
+%   raw_headers(-Headers).
+
+reply_header(Lines, Options) :-
+    option(raw_headers(Headers), Options),
+    !,
+    maplist(string_codes, Headers, Lines).
+reply_header(_, _).
+
+
                  /*******************************
                  *   AUTHORIZATION MANAGEMENT   *
                  *******************************/
@@ -1235,10 +1267,10 @@ rest_(Atom, L, []) :-
 %   If  Authorization  is  the   atom    =|-|=,   possibly   defined
 %   authorization is cleared.  For example:
 %
-%   ==
+%   ```
 %   ?- http_set_authorization('http://www.example.com/private/',
 %                             basic('John', 'Secret'))
-%   ==
+%   ```
 %
 %   @tbd    Move to a separate module, so http_get/3, etc. can use this
 %           too.
@@ -1491,7 +1523,7 @@ update_cookies(Lines, Parts, Options) :-
 %!                     +Options0, -Options) is semidet.
 %
 %   Hook implementation that makes  open_any/5   support  =http= and
-%   =https= URLs for `Mode == read`.
+%   =https= URLs for =|Mode == read|=.
 
 iostream:open_hook(URL, read, Stream, Close, Options0, Options) :-
     (atom(URL) -> true ; string(URL)),
@@ -1636,22 +1668,26 @@ http_close_keep_alive(Address) :-
     forall(get_from_pool(Address, StreamPair),
            close(StreamPair, [force(true)])).
 
-%!  keep_alive_error(+Error)
+%!  keep_alive_error(+Error, +StreamPair)
 %
-%   Deal with an error from reusing  a keep-alive connection. If the
-%   error is due to an I/O error   or end-of-file, fail to backtrack
-%   over get_from_pool/2. Otherwise it is a   real error and we thus
-%   re-raise it.
+%   Deal with an error from  reusing   a  keep-alive  connection. If the
+%   error is due to an I/O error  or end-of-file, fail to backtrack over
+%   get_from_pool/2. Otherwise it is a real   error and we thus re-raise
+%   it. In all cases we close StreamPair rather than returning it to the
+%   pool as we may have done a partial read and thus be out of sync wrt.
+%   the HTTP protocol.
 
-keep_alive_error(keep_alive(closed)) :-
+keep_alive_error(error(keep_alive(closed), _), _) :-
     !,
     debug(http(connection), 'Keep-alive connection was closed', []),
     fail.
-keep_alive_error(io_error(_,_)) :-
+keep_alive_error(error(io_error(_,_), _), StreamPair) :-
     !,
+    close(StreamPair, [force(true)]),
     debug(http(connection), 'IO error on Keep-alive connection', []),
     fail.
-keep_alive_error(Error) :-
+keep_alive_error(Error, StreamPair) :-
+    close(StreamPair, [force(true)]),
     throw(Error).
 
 
@@ -1665,7 +1701,7 @@ keep_alive_error(Error) :-
 %   options based on the the broken-down request-URL.  The following
 %   example redirects all trafic, except for localhost over a proxy:
 %
-%       ==
+%       ```
 %       :- multifile
 %           http:open_options/2.
 %
@@ -1673,7 +1709,7 @@ keep_alive_error(Error) :-
 %           option(host(Host), Parts),
 %           Host \== localhost,
 %           Options = [proxy('proxy.local', 3128)].
-%       ==
+%       ```
 %
 %   This hook may return multiple   solutions.  The returned options
 %   are  combined  using  merge_options/3  where  earlier  solutions
