@@ -701,12 +701,14 @@ postEventWindow(PceWindow sw, EventObj ev)
     goto out;
   }
 
-#pragma GCC diagnostic push
-#pragma GCC diagnostic ignored "-Warray-bounds"
-  if ( sw->focus != (Graphical) sw || notNil(sw->focus_recogniser) )
-  { rval = send(sw, NAME_event, ev, EAV);
-  }
-#pragma GCC diagnostic pop
+  /* This code looks a bit awkward, but prevents a -Warray-bounds
+   * warning from gcc-11
+   */
+  if ( notNil(sw->focus) )
+  { if ( sw->focus != (Graphical) sw || notNil(sw->focus_recogniser) )
+      rval = send(sw, NAME_event, ev, EAV);
+  } else
+    rval = send(sw, NAME_event, ev, EAV);
 
   if ( !rval )
   { ScrollBar sb;
@@ -1729,7 +1731,7 @@ getUserWindow(PceWindow sw)
 
 
 static status
-for_all_tile(TileObj tile, SendFunc f, Any arg)
+for_all_tile(TileObj tile, status (*f)(PceWindow, FrameObj), Any arg)
 { if ( isNil(tile->members) )
     return (*f)(tile->object, arg);
   else
