@@ -3,7 +3,7 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (c)  2016-2021, VU University Amsterdam
+    Copyright (c)  2016-2022, VU University Amsterdam
 			      CWI, Amsterdam
 			      SWI-Prolog Solutions b.v.
     All rights reserved.
@@ -115,6 +115,7 @@ static int	unify_key(ukey_state *state, word key);
 static void	init_ukey_state(ukey_state *state, trie *trie, Word p);
 static void	destroy_ukey_state(ukey_state *state);
 static void	set_trie_clause_general_undefined(Clause cl);
+static void	trie_destroy(trie *trie);
 
 #undef LDFUNC_DECLARATIONS
 
@@ -139,15 +140,18 @@ write_trie_ref(IOSTREAM *s, atom_t aref, int flags)
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 GC a trie. Note that the  Prolog predicate trie_destroy/1 merely empties
-the trie, leaving its destruction to the atom garbage collector.
+the trie, leaving its destruction to   the  atom garbage collector. This
+function is also called from pl-tabling.c   when  we accidentally create
+the same variant trie concurrently and  need   to  dispose  of one. As a
+result, this must be public and thread-safe.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
-static int
+int
 release_trie_ref(atom_t aref)
 { tref *ref = PL_blob_data(aref, NULL, NULL);
-  trie *t;
+  trie *t = ref->trie;
 
-  if ( (t=ref->trie) )
+  if ( t && COMPARE_AND_SWAP_PTR(&ref->trie, t, NULL) )
     trie_destroy(t);			/* can be called twice */
 
   return TRUE;
@@ -173,7 +177,7 @@ load_trie(IOSTREAM *fd)
 
 static PL_blob_t trie_blob =
 { PL_BLOB_MAGIC,
-  PL_BLOB_UNIQUE,
+  0,
   "trie",
   release_trie_ref,
   NULL,
@@ -221,7 +225,7 @@ trie_create(alloc_pool *pool)
 }
 
 
-void
+static void
 trie_destroy(trie *trie)
 { DEBUG(MSG_TRIE_GC, Sdprintf("Destroying trie %p\n", trie));
   trie->magic = TRIE_CMAGIC;
@@ -410,7 +414,7 @@ prune_node(trie *trie, trie_node *n)
     { switch( children.any->type )
       { case TN_KEY:
 	  if ( COMPARE_AND_SWAP_PTR(&p->children.any, children.any, NULL) )
-	    PL_free(children.any);
+	    free_to_pool(trie->alloc_pool, children.key, sizeof(*children.key));
 	  break;
 	case TN_HASHED:
 	  deleteHTable(children.hash->table, (void*)n->key);
@@ -492,7 +496,7 @@ prune_trie(trie *trie, trie_node *root,
       { switch( children.any->type )
 	{ case TN_KEY:
 	    if ( COMPARE_AND_SWAP_PTR(&p->children.any, children.any, NULL) )
-	      PL_free(children.any);
+	      free_to_pool(trie->alloc_pool, children.key, sizeof(*children.key));
 	    break;
 	  case TN_HASHED:
 	    deleteHTable(children.hash->table, (void*)n->key);
@@ -576,7 +580,8 @@ insert_child(DECL_LD trie *trie, trie_node *n, word key)
     { switch( children.any->type )
       { case TN_KEY:
 	{ if ( children.key->key == key )
-	  { return children.key->child;
+	  { destroy_node(trie, new);	/* someone else did this */
+	    return children.key->child;
 	  } else
 	  { trie_children_hashed *hnode;
 
@@ -1790,9 +1795,11 @@ unify_key(DECL_LD ukey_state *state, word key)
 	    bindConst(p, w);
 	  else
 	    return overflowCode(0);
-	} else
+	} else if ( tag(w) == tag(*p) && isIndirect(*p) )
 	{ if ( !equalIndirect(w, *p) )
 	    return FALSE;
+	} else
+	{ return FALSE;
 	}
       }
 
